@@ -212,7 +212,7 @@ namespace dlr {
         message << "First call to dsyev_ returns " << INFO
                 << ".  Something is wrong.";
         DLR_THROW3(ValueException,
-                   "eigenvectorsSymmetric()",
+                   "eigenvaluesSymmetric()",
                    message.str().c_str());
       }
     
@@ -227,10 +227,10 @@ namespace dlr {
       // Check for errors.
       if(INFO != 0L) {
         std::ostringstream message;
-        message << "First call to dsyev_ returns " << INFO
+        message << "Second call to dsyev_ returns " << INFO
                 << ".  Something is wrong.";
         DLR_THROW3(ValueException,
-                   "eigenvectorsSymmetric()",
+                   "eigenvaluesSymmetric()",
                    message.str().c_str());
       }
 
@@ -240,6 +240,139 @@ namespace dlr {
     }
 
 
+    void
+    eigenvectors(const Array2D<Float64>& inputArray,
+                 Array1D< std::complex<Float64> >& eigenvalues,
+                 Array2D< std::complex<Float64> >& eigenvectors,
+                 bool isSortRequired)
+    {
+      // Argument checking.
+      if(inputArray.size() == 0) {
+        DLR_THROW3(ValueException,
+                   "eigenvectors()",
+                   "Argument inputArray cannot have zero size.");
+      }
+      if(inputArray.rows() != inputArray.columns()) {
+        DLR_THROW3(ValueException,
+                   "eigenvectors()",
+                   "Argument inputArray must be square.");
+      }
+    
+      // Transpose A to match LAPACK's convention.
+      size_t dimension = inputArray.rows();
+      Array2D<Float64> aTransposeColumnMajor = inputArray.copy();
+
+      // Set up storage for return values.
+      Array1D<Float64> eigenvaluesReal(dimension);
+      Array1D<Float64> eigenvaluesImag(dimension);
+      Array2D<Float64> leftEigenvectorsTranspose(dimension, dimension);
+      Array2D<Float64> rightEigenvectorsTranspose(dimension, dimension);
+
+      // Set up arguments for the LAPACK call.
+      char JOBVL = 'V'; // Don compute left eigenvectors.
+      char JOBVR = 'N'; // Don't compute right eigenvectors.
+      Int32 N = static_cast<Int32>(dimension);
+      Int32 LDA = static_cast<Int32>(dimension);
+      Int32 LDVL = static_cast<Int32>(dimension);
+      Int32 LDVR = static_cast<Int32>(dimension);
+      Float64 WORK;
+      Int32 LWORK = -1;
+      Int32 INFO;
+
+      // Call once to request optimal workspace size.
+      dgeev_(&JOBVL, &JOBVR, &N, aTransposeColumnMajor.data(), &LDA,
+             eigenvaluesReal.data(), eigenvaluesImag.data(),
+             leftEigenvectorsTranspose.data(), &LDVL,
+             rightEigenvectorsTranspose.data(), &LDVR,
+             &WORK, &LWORK, &INFO);
+
+      // Check for errors.
+      if(INFO != 0L) {
+        std::ostringstream message;
+        message << "First call to dgeev_ returns " << INFO
+                << ".  Something is wrong.";
+        DLR_THROW3(ValueException, "eigenvectors()", message.str().c_str());
+      }
+    
+      // Resize workspace.
+      LWORK = static_cast<Int32>(WORK);
+      Array1D<Float64> doubleWorkSpace(static_cast<size_t>(LWORK));
+
+      // Call again to really compute the eigenvectors.
+      dgeev_(&JOBVL, &JOBVR, &N, aTransposeColumnMajor.data(), &LDA,
+             eigenvaluesReal.data(), eigenvaluesImag.data(),
+             leftEigenvectorsTranspose.data(), &LDVL,
+             rightEigenvectorsTranspose.data(), &LDVR,
+             doubleWorkSpace.data(), &LWORK, &INFO);
+
+      // Check for errors.
+      if(INFO != 0L) {
+        std::ostringstream message;
+        message << "Second call to dgeev_ returns " << INFO
+                << ".  Something is wrong.";
+        DLR_THROW3(ValueException, "eigenvectors()", message.str().c_str());
+      }
+
+      // Check the sizes of the return references.
+      if(eigenvalues.size() != dimension) {
+        eigenvalues.reinit(dimension);
+      }
+      if(eigenvectors.rows() != dimension
+         || eigenvectors.columns() != dimension) {
+        eigenvectors.reinit(dimension, dimension);
+      }
+
+      // Copy eigenvalues and eigenvectors.
+      for(size_t ii = 0; ii < dimension; ++ii) {
+        eigenvalues[ii].real() = eigenvaluesReal[ii];
+        eigenvalues[ii].imag() = eigenvaluesImag[ii];
+        if(eigenvaluesImag[ii] == 0.0) {
+          for(size_t jj = 0; jj < dimension; ++jj) {
+            eigenvectors(jj, ii).real() = leftEigenvectorsTranspose(ii, jj);
+            eigenvectors(jj, ii).imag() = 0.0;
+          }
+        } else {
+          eigenvalues[ii + 1].real() = eigenvaluesReal[ii + 1];
+          eigenvalues[ii + 1].imag() = eigenvaluesImag[ii + 1];
+          for(size_t jj = 0; jj < dimension; ++jj) {
+            eigenvectors(jj, ii).real() =
+              leftEigenvectorsTranspose(ii, jj);
+            eigenvectors(jj, ii).imag() =
+              leftEigenvectorsTranspose(ii + 1, jj);
+            eigenvectors(jj, ii + 1).real() =
+              leftEigenvectorsTranspose(ii, jj);
+            eigenvectors(jj, ii + 1).imag() =
+              -leftEigenvectorsTranspose(ii + 1, jj);
+          }
+          ++ii;
+        }
+      }
+
+      // Do eigenvalues have to be in descending order of magnitude.
+      if(isSortRequired) {
+        Array1D<double> magnitudes2(eigenvalues.size());
+        for(size_t ii = 0; ii < eigenvalues.size(); ++ii) {
+          magnitudes2[ii] = (eigenvalues[ii].real() * eigenvalues[ii].real()
+                             + eigenvalues[ii].imag() * eigenvalues[ii].imag());
+        }
+        Array1D<size_t> outputOrder = argsort(magnitudes2);
+        std::reverse(outputOrder.begin(), outputOrder.end());
+
+        Array1D< std::complex<Float64> > sortedEigenvalues(eigenvalues.size());
+        Array2D< std::complex<Float64> > sortedEigenvectors(
+          eigenvectors.rows(), eigenvectors.columns());
+        for(size_t ii = 0; ii < dimension; ++ii) {
+          sortedEigenvalues[ii] = eigenvalues[outputOrder[ii]];
+          for(size_t jj = 0; jj < dimension; ++jj) {
+            sortedEigenvectors(jj, ii) = eigenvectors(jj, outputOrder[ii]);
+          }
+        }
+        eigenvalues = sortedEigenvalues;
+        eigenvectors = sortedEigenvectors;
+      }
+    }
+
+    
     void
     eigenvectorsSymmetric(const Array2D<Float64>& inputArray,
                           Array1D<Float64>& eigenvalues,
@@ -311,7 +444,7 @@ namespace dlr {
       // Check for errors.
       if(INFO != 0L) {
         std::ostringstream message;
-        message << "First call to dsyev_ returns " << INFO
+        message << "Second call to dsyev_ returns " << INFO
                 << ".  Something is wrong.";
         DLR_THROW3(ValueException,
                    "eigenvectorsSymmetric()",
@@ -628,6 +761,152 @@ namespace dlr {
     }
     
 
+    // This function computes the QR factorization of a general
+    // matrix.
+    void
+    qrFactorization(const Array2D<Float64>& inputArray,
+                    Array2D<Float64>& qArray,
+                    Array2D<Float64>& rArray)
+    {
+      // Argument checking.
+      if(inputArray.size() == 0) {
+        qArray.clear();
+        rArray.clear();
+        return;
+      }
+
+      // Set up scalar arguments for the LAPACK routine.
+      Int32 rows = static_cast<Int32>(inputArray.rows());
+      Int32 columns = static_cast<Int32>(inputArray.columns());
+      Int32 lda = rows;
+      Float64 temporaryWorkspace;
+      Int32 lwork = -1;  // Request info on optimal workspace size.
+      Int32 info;
+
+      // Set up array arguments for the LAPACK routine.
+      Array2D<Float64> AColumnMajor = inputArray.transpose();
+      Array1D<Float64> tauArray = Array1D<Float64>(std::min(rows, columns));
+    
+      // Now invoke the LAPACK routine to find the optimal workspace
+      // size.
+      dgeqrf_(&rows, &columns, AColumnMajor.data(), &lda, tauArray.data(),
+             &temporaryWorkspace, &lwork, &info);
+
+      // Check for errors.
+      if(info != 0L) {
+        std::ostringstream message;
+        message << "First call to dgeqrf_ returns " << info
+                << ".  Something is wrong.";
+        DLR_THROW3(ValueException, "qrFactorization()", message.str().c_str());
+      }
+    
+      // Resize workspace.
+      lwork = static_cast<Int32>(temporaryWorkspace);
+      Array1D<Float64> doubleWorkspace(static_cast<size_t>(lwork));
+
+      // Call again to solve the system of equations.
+      dgeqrf_(&rows, &columns, AColumnMajor.data(), &lda, tauArray.data(),
+             doubleWorkspace.data(), &lwork, &info);
+    
+      // Check for errors.
+      if(info != 0L) {
+        std::ostringstream message;
+        message << "Second call to dgeqrf_ returns " << info
+                << ".  Something is wrong.";
+        DLR_THROW3(ValueException, "qrFactorization()", message.str().c_str());
+      }
+
+      // Prepare reference arguments to receive the result.
+      if((rArray.rows() != inputArray.rows())
+         || (rArray.columns() != inputArray.columns())) {
+        rArray.reinit(inputArray.rows(), inputArray.columns());
+      }
+      rArray = 0.0;
+      if((qArray.rows() != inputArray.rows())
+         || (qArray.columns() != inputArray.rows())) {
+        qArray.reinit(inputArray.rows(), inputArray.rows());
+      }
+      qArray = 0.0;
+
+      // Recover the upper trapezoidal matrix R.
+      size_t numNonzeroRows = std::min(inputArray.rows(), inputArray.columns());
+      for(size_t row = 0; row < numNonzeroRows; ++row) {
+        for(size_t column = row; column < rArray.columns(); ++column) {
+          rArray(row, column) = AColumnMajor(column, row);
+        }
+      }
+
+      // Recover the orthogonal matrix Q.  Here's a quote from the lapack code:
+      //
+      // *  The matrix Q is represented as a product of elementary reflectors
+      // *
+      // *     Q = H(1) H(2) . . . H(k), where k = min(m,n).
+      // *
+      // *  Each H(i) has the form
+      // *
+      // *     H(i) = I - tau * v * v'
+      // *
+      // *  where tau is a real scalar, and v is a real vector with
+      // *  v(1:i-1) = 0 and v(i) = 1; v(i+1:m) is stored on exit in A(i+1:m,i),
+      // *  and tau in TAU(i).
+
+      // Start with identity.
+      for(size_t row = 0; row < qArray.rows(); ++row) {
+        qArray(row, row) = static_cast<Float64>(1.0);
+      }
+      
+      // Now subtract out the tau * v * v^T terms.
+      Array1D<Float64> vArray(qArray.rows());
+      Array2D<Float64> reflector(qArray.rows(), qArray.rows());
+      for(size_t ii = 0; ii < numNonzeroRows; ++ii) {
+        // Get ready to compute the next factor.  Could do this more
+        // efficiently.
+        reflector = 0.0;
+        for(size_t kk = 0; kk < reflector.rows(); ++kk) {
+          reflector(kk, kk) = 1.0;
+        }
+        
+        // Compute tau and v.
+        Float64 tau = tauArray[ii];
+        // We don't need a loop here because the zeros from last
+        // iteration stick around.
+        // 
+        // for(size_t jj = 0; jj < ii; ++jj) {
+        //   vArray[jj] = 0.0;
+        // }
+        if(ii > 0) {
+          vArray[ii - 1] = 0.0;
+        }
+        vArray[ii] = 1.0;
+        for(size_t jj = ii + 1; jj < qArray.rows(); ++jj) {
+          vArray[jj] = AColumnMajor(ii, jj);
+        }
+
+        // Subtract tau * v * v^T.
+        for(size_t row = ii; row < qArray.rows(); ++row) {
+          for(size_t column = ii; column < qArray.columns(); ++column) {
+            reflector(row, column) -= tau * vArray[row] * vArray[column];
+          }
+        }
+
+        qArray = matrixMultiply(qArray, reflector);
+      }
+
+      // Make sure diagonal elements of rArray are non-negative, as
+      // promised.
+      for(size_t ii = 0; ii < inputArray.rows(); ++ii) {
+        if(rArray(ii, ii) == 0.0) {
+          continue;
+        }
+        if(rArray(ii, ii) < 0.0) {
+          rArray *= -1.0;
+          qArray *= -1.0;
+          break;
+        }
+      }
+    }
+
+    
     // This function accepts an Array2D<Float64> instance having at least
     // as many rows as columns, and returns the Moore-Penrose
     // pseudoinverse.
@@ -646,7 +925,9 @@ namespace dlr {
     singularValueDecomposition(const Array2D<Float64>& inputArray,
                                Array2D<Float64>& uArray,
                                Array1D<Float64>& sigmaArray,
-                               Array2D<Float64>& vTransposeArray)
+                               Array2D<Float64>& vTransposeArray,
+                               bool isNullSpaceRequired,
+                               bool isFullRangeRequired)
     {
       // Argument checking.
       if(inputArray.size() == 0) {
@@ -654,7 +935,11 @@ namespace dlr {
                    "singularValueDecomposition()",
                    "Argument inputArray cannot have zero size.");
       }
-    
+
+      // Note(xxx): Fix things so that isFullRangeRequired is not ignored.
+      isFullRangeRequired = isNullSpaceRequired;
+      
+      // 
       // We do not transpose A. Instead, we let LAPACK operate on the
       // untransposed matrix as if it were column major and then swap U
       // and VT at the end.  We make a copy here since the contents of
@@ -665,16 +950,28 @@ namespace dlr {
       size_t numberOfSingularValues =
         std::min(inputArray.rows(), inputArray.columns());
 
-      if((vTransposeArray.rows() != numberOfSingularValues)
+      size_t numberOfVRows;
+      if(isNullSpaceRequired) {
+        numberOfVRows = inputArray.columns();
+      } else {
+        numberOfVRows = numberOfSingularValues;
+      }
+      if((vTransposeArray.rows() != numberOfVRows)
          || (vTransposeArray.columns() != inputArray.columns())) {
-        vTransposeArray.reinit(numberOfSingularValues, inputArray.columns());
+        vTransposeArray.reinit(numberOfVRows, inputArray.columns());
       }
       // Shallow copy.
       Array2D<Float64> uColumnMajor = vTransposeArray;
 
+      size_t numberOfUColumns;
+      if(isFullRangeRequired) {
+        numberOfUColumns = inputArray.rows();
+      } else {
+        numberOfUColumns = numberOfSingularValues;
+      }
       if((uArray.rows() != inputArray.rows())
-         || (uArray.columns() != numberOfSingularValues)) {
-        uArray.reinit(inputArray.rows(), numberOfSingularValues);
+         || (uArray.columns() != numberOfUColumns)) {
+        uArray.reinit(inputArray.rows(), numberOfUColumns);
       }
       // Shallow copy.
       Array2D<Float64> vTColumnMajor = uArray;
@@ -685,7 +982,12 @@ namespace dlr {
       Array1D<Int32> integerWorkSpace(8 * numberOfSingularValues);
     
       // Set up arguments for the LAPACK call.
-      char JOBZ = 'S';  // Compute only min(M, N) columns of U and rows of VT.
+      char JOBZ;
+      if(isNullSpaceRequired) {
+        JOBZ = 'A';
+      } else {
+        JOBZ = 'S';  // Compute only min(M, N) columns of U and rows of VT.
+      }
       Int32 M = static_cast<Int32>(inputArray.columns());
       Int32 N = static_cast<Int32>(inputArray.rows());
       Int32 LDA = M;
@@ -999,7 +1301,7 @@ namespace dlr {
       // Check for errors.
       if(INFO != 0L) {
         std::ostringstream message;
-        message << "First call to dgesdd_ returns " << INFO
+        message << "Second call to dgesdd_ returns " << INFO
                 << ".  Something is wrong.";
         DLR_THROW3(ValueException,
                    "singularValues()",
